"use strict";var b=Object.defineProperty;var g=Object.getOwnPropertyDescriptor;var m=Object.getOwnPropertyNames;var w=Object.prototype.hasOwnProperty;var k=(e,t)=>{for(var s in t)b(e,s,{get:t[s],enumerable:!0})},x=(e,t,s,c)=>{if(t&&typeof t=="object"||typeof t=="function")for(let r of m(t))!w.call(e,r)&&r!==s&&b(e,r,{get:()=>t[r],enumerable:!(c=g(t,r))||c.enumerable});return e};var S=e=>x(b({},"__esModule",{value:!0}),e);var V={};k(V,{initialValue:()=>f,isMap:()=>l,isMapInstance:()=>T,isObjectLiteral:()=>u,isSet:()=>y,isSetInstance:()=>j,merge:()=>M,objectKeys:()=>p});module.exports=S(V);var f={},A=new Set(["__proto__","constructor","prototype"]);function l(e){return T(e)}function T(e){return e instanceof Map}function y(e){return j(e)}function j(e){return e instanceof Set}function u(e){return!!e&&e.constructor===Object}function p(e){return Object.keys(e).filter(t=>!A.has(t))}function M(e,t,...s){let c=f;for(let r of[e,t,...s]){if(!u(r))throw new TypeError("Expected all arguments to be object literals.");let a={...c},O=p(r);for(let i of O){let o=a[i],n=r[i];if(o!==n){if(u(o)&&u(n)){a[i]=M(o,n);continue}if(Array.isArray(o)&&Array.isArray(n)){a[i]=[...new Set([...o,...n])];continue}if(l(o)&&l(n)){a[i]=new Map([...o,...n]);continue}if(y(o)&&y(n)){a[i]=new Set([...o,...n]);continue}a[i]=n}}c=a}return c}0&&(module.exports={initialValue,isMap,isMapInstance,isObjectLiteral,isSet,isSetInstance,merge,objectKeys});
