import * as eslint from 'eslint';

/**
 * Get RuleTester class
 */
declare function getRuleTester(): typeof eslint.RuleTester;
/**
 * Get the prefix of the ruleId used in the rule tester.
 */
declare function getRuleIdPrefix(): string;

export { getRuleIdPrefix, getRuleTester };
