declare const _default: import("eslint").Rule.RuleModule;
export default _default;
export type CustomSpacings = {
    postDelimiter: import("../iterateJsdoc.js").Integer;
    postHyphen: import("../iterateJsdoc.js").Integer;
    postName: import("../iterateJsdoc.js").Integer;
    postTag: import("../iterateJsdoc.js").Integer;
    postType: import("../iterateJsdoc.js").Integer;
};
