/**
 * DON'T EDIT THIS FILE.
 * This file was generated by "scripts/update-lib-configs.js" script.
 */
"use strict"

module.exports = {
    extends: [
        require.resolve("./no-new-in-es2023"),
        require.resolve("./no-new-in-es2022"),
        require.resolve("./no-new-in-es2021"),
        require.resolve("./no-new-in-es2020"),
        require.resolve("./no-new-in-es2019"),
        require.resolve("./no-new-in-es2018"),
        require.resolve("./no-new-in-es2017"),
        require.resolve("./no-new-in-es2016"),
        require.resolve("./no-new-in-es2015"),
        require.resolve("./no-new-in-es5"),
    ],
}
