"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.flow = flow;
function flow(...transforms) {
  return block => transforms.reduce((block, t) => t(block), block);
}
//# sourceMappingURL=index.cjs.map
