/*
	MIT License http://www.opensource.org/licenses/mit-license.php
	Author Tobias Koppers @sokra
*/

"use strict";

// TODO remove in next major
module.exports = require("./ModulesInHierarchicalDirectoriesPlugin");
