import type { Grammar } from './Grammar.js';
export declare const typescriptGrammar: Grammar;
export declare const typescriptNameGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const typescriptNamePathGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const typescriptNamePathSpecialGrammar: import("../parslets/Parslet.js").ParsletFunction[];
